/*
 * Indie Game Jame 2 - example game "framework"
 * 2004/3/15 checker@d6.com
 */
#include "shared/core/first.h"

// windows crap
#define WIN32_LEAN_AND_MEAN
#define WIN32_EXTRA_LEAN
#define _WIN32_WINNT 0x0500
#include <windows.h>
#include <windowsx.h>

// system
#include <assert.h>
#define _USE_MATH_DEFINES
#include <math.h>

// gl stuff
#include <gl/gl.h>
#include "glutils.h"

// physics stuff
#include "shared/physics/2d/2d_body_hull.h"
#include "shared/physics/2d/2d_body_list.h"
#include "shared/physics/2d/2d_body_box.h"
#include "shared/physics/2d/2d_body_sphere.h"
#include "shared/physics/2d/rigid/2d_rigid_sim.h"
#include "shared/physics/2d/rigid/2d_rigid_drag_force.h"
#include "shared/physics/2d/rigid/2d_rigid_gravity_force.h"
#include "shared/physics/2d/rigid/2d_rigid_spring_force.h"
#include "shared/input/joystick.h"

using namespace ajb;
using namespace Physics::D2;
using namespace Physics::D2::Rigid;

// misc
#include "utils.h"
#include <vector>
using namespace std;

// media lib
#include "music.h"
#include "zsound.h"
#include "bitmap.h"

#include "property_system.h"
#include "property_system_editor.h"
#include "editor.h"
#include "../../utils/misc/framerate.h"
#include "../../utils/misc/sim_crap.h"

// cb :
#include "gBase.h"
#include "gUtil.h"
#include "gFloatUtil.h"

/********************************* globals **********************************/
  
int const NumberOfSimSteps = 2;
float const FrameRate = 60.f;

// window/camera stuff
char const AppName[] = "cb1";
int WindowWidth = 800;
int WindowHeight = 600;
bool AppQuit = false;  // stuff this to kill yourself
float FOVY = 45.0f, ZNear = 0.1f, ZFar = 1000.0f;  // whatever, we're 2D
float AspectRatio = float(WindowWidth)/float(WindowHeight); // updated in Resize
float CameraAngleRadians = 0.0, CameraX = 0.0, CameraY = 0.0, CameraZ = 100.0;
Vec2 WorldFromWindowCoords( Vec2 window );
Vec2 WindowFromWorldCoords( Vec2 world );

// win32 stuff
HWND WindowHandle;
HDC WindowDC;

// gl stuff
HGLRC WindowRC;

// functions for you to play in
void Resize( int w, int h );
void Initialize( void );
void Uninitialize( void );
void Idle( void );
LONG FAR PASCAL AppWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam);

// system
Music music;

// physics stuff
Sim simulator;
Drag_Force    global_drag(.05f, .05f);    //a little drag helps with stability
Gravity_Force gravity(Vec2(0.0f, -1.0f));
Rigid_Body    world_body(make_Body_Null(),Vec2(0,0),0,0,0);

char *TestLevelFileName = "bodies.physdb";

PROP_PARSE_THIS_FILE;
PROP_STRUCT struct global_properties
{
    float Gravity;
    float DragLinear;
    float DragAngular;
    
    float MouseSpringKs;
    float MouseSpringKd;
    float MouseSpringRestLength;
    
    int	NumControls;
    int	CurControl;
};
static global_properties Globals;
static property *GlobalProperties;
static property_system_editor *GlobalsEditor;

static bool MovingCamera = false;

PROP_STRUCT enum cbControlInput
{
	eLeftStickH,
	eLeftStickV,
	eRightStickH,
	eRightStickV,
	eLeftStickAngle,
	eRightStickAngle,
	eButton_X,
	eButton_O,
	eButton_Square,
	eButton_Triangle,
	eMouseH,
	eMouseV,
	eKey,
	eNull
};

PROP_STRUCT enum cbControlType
{
	eSpringLength,			// spring name , low & high values
	eAngle,					// angle-spring name
};

PROP_STRUCT enum cbControlOrder
{
	eControlValue,	// stick value sets controlled value
	eControlDelta	// stick value adds to controlled value
};

PROP_STRUCT struct cbControlOptions
{
	cbControlInput	m_input;
	cbControlType	m_type;
	cbControlOrder	m_order;
	float			m_controlDeltaSpeed;
	float			m_controlDeltaLo;
	float			m_controlDeltaHi;
	float			m_controlLo;
	float			m_controlRest;
	float			m_controlHi;
	std::string		m_name;
	int				m_gamepad;
	std::string		m_key;
//	char *			m_name;
	
	cbControlOptions();
};

cbControlOptions defaultOptions;

cbControlOptions::cbControlOptions() :
		m_input(eLeftStickH),
		m_type(eSpringLength),
		m_order(eControlValue),
		m_controlDeltaSpeed(1.f),
		m_controlDeltaLo(-FLT_MAX),
		m_controlDeltaHi(FLT_MAX),
		m_controlLo(0.f),
		m_controlRest(1.f),
		m_controlHi(10.f),
		m_name("force0000"),
		m_gamepad(0)
{
}

static property *cbControlProperties;
property_system_editor * cbControlEditor;

struct cbControl
{
	cbControlOptions	m_options;
	float				m_value;
	
	cbControl() :
		m_options(),
		m_value(0.f)
	{
	}
	
	void Tick();
};

// game stuff
struct game_object 
{
    
    game_object( char *texture_file , Rigid_Body *rbInit ) :
            rb(rbInit)
        {
            assert(rb);

            simulator.add(*rb);
        }

    ~game_object( void )
        {
            delete rb;
        }
    
    Rigid_Body *rb;
};

bool my_collision_test_callback(const Body& a, const Body& b,void* rb_a, void* rb_b)
{
//	Rigid_Body * rba = (Rigid_Body *)rb_a;
//	Rigid_Body * rbb = (Rigid_Body *)rb_b;

	if ( a.m_collision_plane == -1 || b.m_collision_plane == -1 )
		return true;
	
	if ( a.m_collision_plane == b.m_collision_plane )
		return false;
	
	return true;
}

static vector<game_object*> objs;
static vector<Binary_Force *> forces;
static vector<cbControl> controls;

void MakeControls();
void TickControls();
void SaveControls();
void LoadControls();

static Vec2 window_mouse, world_mouse;
static Vec2 delta_window_mouse,delta_world_mouse;
static Joystick PS2Controller;

static Binary_Force *spring = 0;

//! Send printf-formatted output to the log, using the eGeneric message type.
//  Same as calling Printf(eGeneric, fmt, ...)
void cbLog(const char* fmt, ...)
{
	va_list			ArgPtr;
	char			TempStr[4096];

	va_start(ArgPtr, fmt);
	_vsnprintf(TempStr, 4096, fmt, ArgPtr);
	va_end(ArgPtr);

	OutputDebugString(TempStr);
	printf(TempStr);
}

GLuint
LoadTexture(char *Filename)
{
    GLuint texture;
    
    Bitmap_Load_Result result;
    verify(LoadBitmapByExtension(Filename,&result));

    // gl stuff
    texture = gl_texture_from_bitmap(result.data, result.width, result.height,
                                     result.bytes_per_pixel);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

    // delete the bitmap
    delete [] result.data;

    return(texture);
}

void
glBody1v(Body *body, bool Solid)
{
    if(body)
    {
		{
			Body_Hull *p = dynamic_cast<Body_Hull*>(body);
			if(p)
			{
				glBegin(Solid ? GL_TRIANGLE_FAN : GL_LINE_LOOP);
				for(int i = 0;i<p->num_verts();++i)
				{
					Vec2 worldspace = p->vert(i);
					glVertex2f(worldspace.x(),worldspace.y());
				}
				glEnd();
			}
		}

		{
			Body_Box *p = dynamic_cast<Body_Box*>(body);
			if(p)
			{
				glBegin(Solid ? GL_TRIANGLE_FAN : GL_LINE_LOOP);
				for(int i = 0;i<p->num_verts();++i)
				{
					Vec2 worldspace = p->vert(i);
					glVertex2f(worldspace.x(),worldspace.y());
				}
				glEnd();
			}
		}

		{
			Body_Sphere *p = dynamic_cast<Body_Sphere*>(body);
			if(p)
			{
				glBegin(Solid ? GL_TRIANGLE_FAN : GL_LINE_LOOP);
				int total_steps = 16;
				Vec2 center = p->x();
				if(Solid)
				{
					glVertex2f(center.x(), center.y());
				}
				for(int a = 0;a<total_steps;++a)
				{
					float Angle = 2.0f * 3.14159265358979323846f * ((float)a / (float)(total_steps - 1));
					float Radius = p->r();
					Vec2 arm(Radius*cos(Angle), Radius*sin(Angle));
					arm += center;
					glVertex2f(arm.x(), arm.y());
				}
				glEnd();
			}
		}

        Body_List *l = dynamic_cast<Body_List*>(body);
        if(l)
        {
            {for(int Index = 0;
                 Index < l->num_bodies();
                 ++Index)
            {
                glBody1v(&l->body(Index), Solid);
            }}
        }
    }
}

void
glForce1v(Binary_Force *Force)
{
    if(Force)
    {
        glBegin(GL_LINES);
        glVertex2f(Force->m_a.x_worldS().x(), Force->m_a.x_worldS().y());
        glVertex2f(Force->m_b.x_worldS().x(), Force->m_b.x_worldS().y());
        glEnd();
    }
}

void
ClearMouseSpring(void)
{
    if(spring)
    {
        simulator.remove(*spring);
        delete spring;
        spring = 0;
    }
}

void
ClearAll(void)
{
    ClearMouseSpring();
    
    for(int i = 0;i < objs.size();++i)
    {
        assert(objs[i]);
        simulator.remove(*objs[i]->rb);
    }

    for(int i = 0;i < forces.size();++i)
    {
        assert(forces[i]);
        simulator.remove(*forces[i]);
    }

    forces.clear();
    objs.clear();
}

const char * stristr(const char * search_in,const char * search_for)
{
	while ( *search_in )
	{
		if ( toupper(*search_in) == toupper(*search_for) )
		{
			const char * in_ptr = search_in + 1;
			const char * for_ptr= search_for + 1;
			while ( *for_ptr && toupper(*for_ptr) == toupper(*in_ptr) )
			{
				for_ptr++; in_ptr++;
			}
			if ( ! *for_ptr) 
				return search_in;
		}
		search_in++;
	}

	return NULL;
}

void
LoadTestFile(void)
{
    int num_rb=0, num_force=0;
    Rigid_Body **rb_list;
    Binary_Force **force_list;

	float scale = 1.f/32.f;
	float density_scale = 1/16.f;

    loadPhysicsFile(TestLevelFileName, 0,22, scale, density_scale,
                    &rb_list, &num_rb, &force_list, &num_force, &world_body);

    for (int i=0; i < num_rb; ++i)
    {
		if ( stristr(rb_list[i]->name().c_str(),"ignore") != NULL )
			continue;		
    
        objs.push_back(new game_object("content/bitmaps/test.jpg",
                                       rb_list[i]));
    }

    for (int i=0; i < num_force; ++i)
    {
		if ( stristr(force_list[i]->name().c_str(),"ignore") != NULL )
			continue;	
			
        forces.push_back(force_list[i]);
        simulator.add(*force_list[i]);
    }
}

/******************************** Initialize ********************************/
/* the window and graphics subsystem are set up but still hidden            */
void Initialize( void )
{
	PS2Controller.init();
		
    InitializeFrameRate(FrameRate);
  
	g_collision_test_callback = my_collision_test_callback;

    Globals.Gravity = -30.0f;
	Globals.DragLinear = 0.01f;
	Globals.DragAngular = 0.01f;

    Globals.MouseSpringKs = 128.0f;
    Globals.MouseSpringKd = 32.0f;
    Globals.MouseSpringRestLength = 0.0f;

    Globals.NumControls = 1;
    Globals.CurControl = 0;
    
    GlobalProperties = CloneProperty(PROP_LOCAL_STRUCT(global_properties));
    ConvertStructToProperty(PROP_LOCAL_STRUCT(global_properties), &Globals,
                            GlobalProperties);
    GlobalsEditor = OpenPropertySystemEditor("Globals", 850, 50, 400, 300);
    EditPropertiesInPlace(GlobalsEditor, GlobalProperties);
    
	cbControlProperties = CloneProperty(PROP_LOCAL_STRUCT(cbControlOptions));
    ConvertStructToProperty(PROP_LOCAL_STRUCT(cbControlOptions), &defaultOptions,
                            cbControlProperties);
    cbControlEditor = OpenPropertySystemEditor("cbControlOptions", 850, 350, 400, 400);
    EditPropertiesInPlace(cbControlEditor, cbControlProperties);
    
    GetTime(); // throw away initial 0.0 so we're seconds-from-startup

    // gl stuff
    glClearColor(1.0f,1.0f,1.0f,0.0f);
    glEnable(GL_CULL_FACE);
    glEnable(GL_NORMALIZE);
    glEnable(GL_DEPTH_TEST);
    glShadeModel(GL_SMOOTH);

    // init our view
    Resize(WindowWidth,WindowHeight);

    // music init
    music.init();
    music.set_subdirectory("content/music/");
    music.setupnotes();
    music.loadseg("background", "simple_loop.sgt", MUSIC_TRACK_MAIN);

    // sound init
    zsoundInit(WindowHandle);

    // initialize the physics
    simulator.add(global_drag);
    simulator.add(gravity);
    simulator.add(world_body);

    LoadTestFile();
    
    LoadControls();
}

bool paused = false;

void UpdateControlCounts()
{
	if ( Globals.NumControls < 1 )
	{
		Globals.NumControls = 1;
	}
	if ( Globals.NumControls != controls.size() )
	{
		controls.resize( Globals.NumControls );
	}
	Globals.CurControl = gClamped(Globals.CurControl,0,Globals.NumControls-1);
}

void RefreshEditors()
{
    ConvertStructToProperty(PROP_LOCAL_STRUCT(global_properties), &Globals,
                            GlobalProperties);
	UpdateDisplay(GlobalsEditor);
    ConvertStructToProperty(PROP_LOCAL_STRUCT(cbControlOptions), &controls[Globals.CurControl],
                            cbControlProperties);	
	UpdateDisplay(cbControlEditor);
}

void Reset()
{
    ClearAll();
    LoadTestFile();
    
    for(int i=0;i<controls.size();i++)
    {
		controls[i].m_value = 0.f;
    }
}
                
/*********************************** Idle ***********************************/
/* called by the message loop, and by paint messages                        */
void Idle( void )
{
	// first copy cur Control to its struct :
	if ( Globals.CurControl >= 0 && Globals.CurControl < controls.size() )
	{
		ConvertPropertyToStruct(FindSubPropertyByName(LocalTemplates, "cbControlOptions"),
								cbControlProperties,
								&controls[Globals.CurControl]);		
	}


	// update from properties :
    ConvertPropertyToStruct(FindSubPropertyByName(LocalTemplates, "global_properties"),
                            GlobalProperties,
                            &Globals);

	UpdateControlCounts();

	// now look at this control :
	RefreshEditors();

	gravity.m_g.set_y(Globals.Gravity);
	global_drag = Drag_Force(Globals.DragLinear,Globals.DragAngular);
    
	PS2Controller.sample();


	/**************************** step the simulator ****************************/   


	if (PS2Controller.button_went_down(Joystick::BUTTON_START,0) ||
		PS2Controller.button_went_down(Joystick::BUTTON_START,1) )
	{
		Reset();
	}
	
	if ( ! paused )
	{
		TickControls();

		if (spring)                             //update
			spring->m_a = Handle(world_body, world_mouse);

		for(int StepCounter = 0;StepCounter < NumberOfSimSteps;++StepCounter) {
			// step as many times as you like
			simulator.step(1/(NumberOfSimSteps*GetFrameRate()));
			// do any game stuff you want here (like per-step collision response, etc.)
		}
		// now wait for the rest of the time so we're consistent on fast machines
		WaitFrameRate(GetFrameRate());
	}
	
    /**************************** render everything *****************************/
   
    // lots of gl stuff
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glDisable(GL_DEPTH_TEST);

    glLoadIdentity();
    glRotated(180.0*(-CameraAngleRadians/M_PI),0,0,1);
    glTranslatef(-CameraX,-CameraY,-CameraZ);

    // draw axes
    glBegin(GL_LINES);
    glColor3f(1,0,0);  glVertex2f(0,0);  glVertex2f(1,0);
    glColor3f(0,1,0);  glVertex2f(0,0);  glVertex2f(0,1);
    glEnd();

    // draw mouse
    glPointSize(4.0);
    glBegin(GL_POINTS);
    glColor3f(0,0,0);  glVertex2fv(world_mouse.m_v);
    glEnd();

    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	
	// iterate all the bodies, rendering them
    glColor3f(0,1,0);
    for(int i = 0;i < objs.size();++i) 
	{
        assert(objs[i]);
        game_object &obj = *objs[i];

        glBody1v(obj.rb->m_Pbody.get(), true);
    }

    glColor3f(0,0,0);
    for(int i = 0;i < objs.size();++i) 
	{
        assert(objs[i]);
        game_object &obj = *objs[i];

        glBody1v(obj.rb->m_Pbody.get(), false);
    }
    
    // iterate all the forces, rendering them
    glColor3f(1,0,0);
    for(int i = 0;i < forces.size();++i)
    {
        assert(forces[i]);
        Binary_Force *Force = forces[i];

        glForce1v(Force);
    }
    glForce1v(spring); 

    glDisable(GL_BLEND);

	/*
	glColor3f(0, 0, 1);
	glBegin(GL_LINES);
	float StickScale = 10.0f;
	glVertex2f(0, 0);
	glVertex2f(StickScale * LeftStick.x(), StickScale * LeftStick.y());
	glVertex2f(0, 0);
	glVertex2f(StickScale * RightStick.x(), StickScale * RightStick.y());
	glEnd();
	*/

    glColor3f(0, 0, 0);
    char buf[100];
    sprintf(buf,"world: %5.1f, %5.1f  window: %5.1f, %5.1f",
            world_mouse.x(),world_mouse.y(),window_mouse.x(),window_mouse.y());
    DrawString(50,50,buf);
    
	Vec2 LeftStick = PS2Controller.left_stick(0);
	Vec2 RightStick = PS2Controller.right_stick(0);
    sprintf(buf,"left: %5.1f, %5.1f  right: %5.1f, %5.1f",
            LeftStick.x(),LeftStick.y(),RightStick.x(),RightStick.y());
    DrawString(50,20,buf);

	char *ButtonStrings[] =
	{
		"BUTTON_RIGHT_D",
		"BUTTON_RIGHT_U",
		"BUTTON_RIGHT_L",
		"BUTTON_RIGHT_R",
		"BUTTON_SHOULDER_R1",
		"BUTTON_SHOULDER_R2",
		"BUTTON_SHOULDER_L1",
		"BUTTON_SHOULDER_L2",
		"BUTTON_RSTICK",
		"BUTTON_LSTICK",
		"BUTTON_START",
		"BUTTON_SELECT",
	};

	int Y = 70;
	for(int ButtonIndex = 0;
		ButtonIndex < Joystick::NUM_BUTTONS;
		++ButtonIndex)
	{
		if(PS2Controller.button_is_down(ButtonIndex))
		{
			DrawString(50.f,(float)Y,ButtonStrings[ButtonIndex]);
			Y += 10;
		}
	}

	draw_obj_attribs(simulator,window_mouse,world_mouse);

    DrawTimingBars();
    SwapBuffers(WindowDC);
    
}

/******************************** AppWndProc ********************************/
/* lame to expose this, but the alternatives all suck too                   */
LONG FAR PASCAL AppWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
    switch (msg)
    {
        // keyboard
        case WM_CHAR:
            break;

        case WM_MOUSEWHEEL:
        {
            float d = float((short)HIWORD(wParam))/float(WHEEL_DELTA);
            CameraZ *= 1.0f + d/10.0f;
            world_mouse = WorldFromWindowCoords(window_mouse);
            break;
        }
        case WM_LBUTTONDOWN:
        {
			SetCapture(hwnd);

            Vec2 body_space_position;
            Rigid_Body* rb = simulator.pick(body_space_position, world_mouse);
            if (rb)
            {
                spring = new Spring_Force(
                    Handle(world_body, world_mouse),
                    Handle(*rb, body_space_position), 
                    Globals.MouseSpringKs,
                    Globals.MouseSpringKd,
                    Globals.MouseSpringRestLength
                    );
                simulator.add(*spring);
            }
        } break;

        case WM_LBUTTONUP:
        {
			if(!(wParam & MK_MBUTTON & MK_RBUTTON))
			{
				ReleaseCapture();
			}

            ClearMouseSpring();
        } break;

        case WM_RBUTTONDOWN:
        {
			SetCapture(hwnd);

			if(spring)
            {
                forces.push_back(spring);
                spring = 0;
            }
            break;
        }
        case WM_MBUTTONDOWN:
        {
            MovingCamera = true;
        } break;
        case WM_MBUTTONUP:
        {
            MovingCamera = false;
        } break;

        case WM_RBUTTONUP:
        {
			if(!(wParam & MK_MBUTTON & MK_LBUTTON))
			{
				ReleaseCapture();
			}
        } break;
            
        case WM_KEYDOWN: {
            switch(wParam) {
				case ' ':
					paused = !paused;
					break;
				case 'S':
				case 's':
					SaveControls();
					break;
				case 'L':
				case 'l':
					LoadControls();
					break;
				
                case VK_LEFT: {
                    CameraAngleRadians += float((5.0/180.0)*M_PI);
                    break;
                }
                case VK_RIGHT: {
                    CameraAngleRadians -= float((5.0/180.0)*M_PI);
                    break;
                }

                case 'M': {
                    // demo the audio stuff
                    if(0x8000 & GetKeyState(VK_SHIFT)) {
                        music.play("background");
                    } else {
                        zsoundPlay("content/sounds/bad.0002.wav");
                    }
                    break;
                }

				case 'R': {
					Reset();
                } break;     

                case 'E': {
                    launchEditor(TestLevelFileName);
                    ClearAll();
                    LoadTestFile();
                } break;
            
                case VK_ESCAPE: {
                    AppQuit = true;
                    break;
                }
            }
            break;
        }
        case WM_SYSKEYDOWN:
        case WM_SYSKEYUP:
        case WM_KEYUP:
            break;

        case WM_MOUSEMOVE: {
			Vec2 new_window_mouse = Vec2(float(GET_X_LPARAM(lParam)),float(GET_Y_LPARAM(lParam)));
            Vec2 new_world_mouse = WorldFromWindowCoords(new_window_mouse);
			delta_window_mouse = new_window_mouse - window_mouse;
            delta_world_mouse = new_world_mouse - world_mouse;
			if(MovingCamera)
			{
				CameraX -= delta_world_mouse.x();
				CameraY -= delta_world_mouse.y();
			}
            window_mouse = new_window_mouse;
            world_mouse = new_world_mouse;
            break;
        }

         
            /**************** below is stuff you probably don't care about ****************/
         
        case WM_DESTROY: {
            PostQuitMessage(0);
            break;
        }
        case WM_PAINT: {
            PAINTSTRUCT ps;
            BeginPaint(hwnd,&ps);
            Idle();
            EndPaint(hwnd,&ps);
            return 0L;
        }
        case WM_SIZE: {
            Resize(LOWORD(lParam),HIWORD(lParam));
            break;
        }
    }
    return (LONG)DefWindowProc(hwnd,msg,(LONG)wParam,(LONG)lParam);
}

/********************************** Resize **********************************/
/* called on WM_SIZE, and at the bottom of initialize                       */
void Resize( int w, int h )
{
    // compute our frustrum planes, the x,y plane values are on the near clip plane
    WindowWidth = w;
    WindowHeight = h;
    AspectRatio = float(w)/float(h);
    float ymax = ZNear * float(tan(FOVY/2.0));
    float ymin = -ymax;
    float xmin = AspectRatio * ymin;
    float xmax = AspectRatio * ymax;
   
    // load the viewport and projection matrix, gl stuff
    glViewport(0,0,w,h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    // gluPerspective doesn't set the near plane as the projection
    // plane, so you can't do good math, so do it ourselves
    glFrustum(xmin,xmax,ymin,ymax,ZNear,ZFar);
    glMatrixMode(GL_MODELVIEW);
}

/******************************* Uninitialize *******************************/
/* called before the graphics is destroyed, but the window is already toast */
void Uninitialize( void )
{
}

/********************************* WinMain **********************************/

int PASCAL WinMain(HINSTANCE hInst, HINSTANCE hPrev, LPSTR szCmdLine, int sw)
{
    // register and create our window
    WNDCLASS cls;
    if(!hPrev) {
        cls.hCursor        = LoadCursor(NULL,IDC_ARROW);
        cls.hIcon          = 0;
        cls.lpszMenuName   = 0;
        cls.lpszClassName  = AppName;
        cls.hbrBackground  = 0;
        cls.hInstance      = hInst;
        cls.style          = CS_VREDRAW | CS_HREDRAW | CS_DBLCLKS | CS_OWNDC;
        cls.lpfnWndProc    = (WNDPROC)AppWndProc;
        cls.cbWndExtra     = 0;
        cls.cbClsExtra     = 0;

        if (!RegisterClass(&cls))
            return FALSE;
    }

    DWORD Style = WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS;
    RECT WindowRect = { 0, 0 };
    WindowRect.right = WindowWidth;
    WindowRect.bottom = WindowHeight;
    BOOL Menu = FALSE;

    AdjustWindowRect(&WindowRect,Style,Menu);
    
    WindowHandle = CreateWindow(AppName,AppName,Style,
                                0,50,
                                WindowRect.right-WindowRect.left,
                                WindowRect.bottom-WindowRect.top,
                                0,0,hInst,0);

    // create our gl stuff
    WindowDC = GetDC(WindowHandle);
    PIXELFORMATDESCRIPTOR FormatDescriptor = {
        sizeof (PIXELFORMATDESCRIPTOR), 1,
        PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
        PFD_TYPE_RGBA, 32,   // 32bpp
        0,0, 0,0, 0,0, 0,0,
        0, 0,0,0,0,
        24, 8,               // depth, stencil
        0,
        PFD_MAIN_PLANE,
        0,
        0,0,0
    };
    int FormatIndex = ChoosePixelFormat(WindowDC,&FormatDescriptor);
    SetPixelFormat(WindowDC,FormatIndex,&FormatDescriptor);
    WindowRC = wglCreateContext(WindowDC);
    wglMakeCurrent(WindowDC,WindowRC);

    // okay, ready to go
   
    Initialize();    // call the app
   
    ShowWindow(WindowHandle,sw);
   
    // pump the message queue
    MSG msg;
    while(!AppQuit) {
        if(PeekMessage(&msg,0,0,0,PM_REMOVE)) {
            if(msg.message == WM_QUIT)
                break;
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        } else {
            Idle();    // call the app
        }
    }

    Uninitialize();  // call the app
   
    // clean up gl stuff and exit
    wglMakeCurrent(0,0);
    wglDeleteContext(WindowRC);
    ReleaseDC(WindowHandle,WindowDC);
   
    return int(msg.wParam);
}

/********************** utilities that use our globals **********************/
Vec2 WorldFromWindowCoords( Vec2 window )
{
    float xn = float(window.x())/WindowWidth - 0.5f;
    float yn = float(WindowHeight-window.y())/WindowHeight - 0.5f;
    float xc = float(xn*2.0*CameraZ*tan(FOVY/2)*AspectRatio);
    float yc = float(yn*2.0*CameraZ*tan(FOVY/2));
    return Vec2(CameraX,CameraY) + Matrix22::rotation(CameraAngleRadians) * Vec2(xc,yc);
}
Vec2 WindowFromWorldCoords( Vec2 world )
{
    Vec2 c = Matrix22::rotation(-CameraAngleRadians) * (world - Vec2(CameraX,CameraY));
    float xn = float(c.x()/(2.0*CameraZ*tan(FOVY/2)*AspectRatio));
    float yn = float(c.y()/(2.0*CameraZ*tan(FOVY/2)));
    float xs = WindowWidth*(xn + 0.5f);
    float ys = WindowHeight - WindowHeight*(yn + 0.5f);
    return Vec2(xs,ys);
}

float StickScalar(float x,cbControlOrder order)
{
	return fclamp(x,-1.f,1.f);
}

float MouseScalar(float v,float d,cbControlOrder order)
{
	switch(order)
	{
	case eControlValue:
		// -1 to 1 to [0,1]
		// @@ scale
		return v;
	case eControlDelta:
		// -1 to 1 is good
		// @@ scale
		return d;
	}	
	return 0.f;
}

float ButtonScalar(bool b)
{
	if ( b )
		return 1.f;
	else
		return 0.f;
}

float StickAngleScalar(Vec2 v,float offset,float last)
{
	if ( v.x() == 0.f && v.y() == 0.f )
		return last;

	float a = atan2f(v.y(),-v.x()) + offset;
	
	while ( a < (last - PI) )
		a += 2*PI;
	while ( a > (last + PI) )
		a -= 2*PI;
	
	return a;
}

bool IsKeyDown(const int vk)
{
	if (GetAsyncKeyState(vk) & 0x8000)
		return true;

	return false;
}

float cbInput(const cbControl & control)
{
	const cbControlOptions & options = control.m_options;
	
	Vec2 LeftStick = PS2Controller.left_stick(options.m_gamepad);
	Vec2 RightStick = PS2Controller.right_stick(options.m_gamepad);
	
	switch(options.m_input)
	{
	case eLeftStickV:
		return StickScalar(LeftStick.y(),options.m_order);
		break;
	case eLeftStickH:
		return StickScalar(LeftStick.x(),options.m_order);
		break;
	case eRightStickV:
		return StickScalar(RightStick.y(),options.m_order);
		break;
	case eRightStickH:
		return StickScalar(RightStick.x(),options.m_order);
		break;
	case eLeftStickAngle:
		return StickAngleScalar(LeftStick,options.m_controlRest,control.m_value);
		break;
	case eRightStickAngle:
		return StickAngleScalar(RightStick,options.m_controlRest,control.m_value);
		break;
	case eButton_X:
		return ButtonScalar( PS2Controller.button_is_down(0,options.m_gamepad) );
		break;
	case eButton_O:
		return ButtonScalar( PS2Controller.button_is_down(1,options.m_gamepad) );
		break;
	case eButton_Square:
		return ButtonScalar( PS2Controller.button_is_down(2,options.m_gamepad) );
		break;
	case eButton_Triangle:
		return ButtonScalar( PS2Controller.button_is_down(3,options.m_gamepad) );
		break;
	case eMouseV:
		return MouseScalar(window_mouse.y(),delta_window_mouse.y(), options.m_order);
		break;
	case eMouseH:
		return MouseScalar(window_mouse.x(),delta_window_mouse.x(), options.m_order);
		break;
	case eKey:
	{
		char key = options.m_key[0];
		return ButtonScalar( IsKeyDown(key) );
		break;
	}
	case eNull:
		break;
	};
	
	return 0.f;
}

void cbControl::Tick()
{
	// find the Binary_force

	if ( m_options.m_name.empty() )
		return; // null

	Binary_Force * force = NULL;
	for(int i=0;i<forces.size();i++)
	{
		if ( forces[i]->name() == m_options.m_name )
		{
			force = forces[i];
			break;
		}
	}
	if ( force == NULL )
	{
		cbLog("couldn't hook up to force : %s\n",m_options.m_name);
		return;
	}
	
	float input = cbInput(*this);

	float value;
	
 	switch(m_options.m_order)
	{
		case eControlValue:
		{
			if ( m_options.m_input == eLeftStickAngle || m_options.m_input == eRightStickAngle )
			{
				value = input;
			}
			else
			{
				if ( input < 0.f )
					value = flerp(m_options.m_controlRest,m_options.m_controlLo,-input);
				else
					value = flerp(m_options.m_controlRest,m_options.m_controlHi,input);
			}
			break;	
		}
		case eControlDelta:
		{
			value = m_value + m_options.m_controlDeltaSpeed * input;
			value = fclamp(value,m_options.m_controlDeltaLo,m_options.m_controlDeltaHi);
			break;
		}
	}
	
	m_value = value;
	
	Spring_Force * sf = dynamic_cast<Spring_Force *>(force);
	Angle_Spring_Force * asf = dynamic_cast<Angle_Spring_Force *>(force);
	//Pt2Pt_Constraint * p2p = dynamic_cast<Pt2Pt_Constraint *>(force);
	
	//cbLog(" Value : %s\n",value );
	
	switch(m_options.m_type)
	{
	case eSpringLength:
		if ( sf )
			sf->m_rest_len = value;
		else
			cbLog("Error : Not a Spring : %s\n",m_options.m_name);
		break;
	case eAngle:
		if ( asf )
			asf->m_rest_len = value;
		else
			cbLog("Error : Not an Angle Spring : %s\n",m_options.m_name);
		break;
	
	}
	
}

void TickControls()
{
	for(int i=0;i<controls.size();i++)
	{
		controls[i].Tick();
	}
}

char * Filename = "controls.prop";

void SaveControls()
{
	vector<property *> temps;

	property * Parent = CreateProperty(SubGroupProp,"Parent");

	{
		property * prop = CloneProperty(GlobalProperties);
		SetPropertyGroup(prop,Parent);
	}

	for(int i=0;i<controls.size();i++)
	{
		property * prop = CloneProperty(PROP_LOCAL_STRUCT(cbControlOptions));
		ConvertStructToProperty(PROP_LOCAL_STRUCT(cbControlOptions), 
								&controls[i].m_options,
								prop);
                            
		SetPropertyGroup(prop,Parent);
		
		temps.push_back(prop);
	}

	WritePropertyToFile(Filename,Parent);
	
	// @@ ??
	DeleteProperty( Parent );
	// bad leak, but who cares
}

void LoadControls()
{
	property * Read = ReadPropertyFromFile(Filename);
	
	property * First = FirstSubProperty(Read);
	
    ConvertPropertyToStruct(FindSubPropertyByName(LocalTemplates, "global_properties"),
                            First,
                            &Globals);

	UpdateControlCounts();

	property * Next = First;
	
	for(int i=0;i<controls.size();i++)
	{
		Next = NextProperty(Next);
		if ( Next == NULL )
			break;

		ConvertPropertyToStruct(FindSubPropertyByName(LocalTemplates, "cbControlOptions"),
                            Next,
                            &controls[i].m_options);
		
	}
	
	RefreshEditors();
}
