/*

 Only very generic things go here; in particular, things that
	are needed even in minimal files, or things that must come very early
*/

#pragma once

//-------------------------------------------------------------------

#pragma warning(disable : 4514) // unreferenced inline function
#pragma warning(disable : 4710) // inline function not inlined
#pragma warning(disable : 4711) // function selected for inlining
#pragma warning(disable : 4714) // forceinline not inlined
#pragma warning(disable : 4127) // conditional is constant
#pragma warning(disable : 4100) // unreferenced parameter
#pragma warning(disable : 4786) // symbol longer than 256 chars
#pragma warning(disable : 4097) // using typedef as synonym for class name

//-------------------------------------------------------------------

// define NULL here so that stdlib.h isn't necessary
#ifndef NULL
#define NULL 0
#endif

// unsigned types.
typedef unsigned char   	ubyte;
typedef unsigned short		uword;
typedef unsigned long		ulong;

typedef __int64        		int64;
typedef unsigned __int64	uint64;

//-------------------------------------------------------------------

//! Silences compiler warning about unused parameters/variables
#define UNUSED_PARAMETER(x)		(x)
#define REFERENCE_TO_VARIABLE	UNUSED_PARAMETER

//! Disallows the compiler defined default ctor
#define FORBID_DEFAULT_CTOR(x) x()

//! Disallows the compiler defined copy ctor
#define FORBID_COPY_CTOR(x)    x(const x&)

//! Disallows the compiler defined assignment operator
#define FORBID_ASSIGNMENT(x)   void operator=(const x&)

#define FORBID_CLASS_STANDARDS(x)	\
	FORBID_ASSIGNMENT(x);	\
	FORBID_COPY_CTOR(x)	

#define MAKE_COMPARISONS_FROM_LESS_AND_EQUALS(this_type)	\
	bool operator <= (const this_type & other) const { return ! ( other < *this ); } \
	bool operator >= (const this_type & other) const { return ! ( *this < other ); } \
	bool operator >  (const this_type & other) const { return ( other < *this ); }	\
	bool operator != (const this_type & other) const { return ! ( *this == other ); }

#define STACK_ARRAY(name,type,count)	type * name = (type *) _alloca(sizeof(type)*(count));

//-------------------------------------------------------------------

#define _Stringize( L )			#L
#define _DoMacro1( M, X )		M(X)
#define _DoMacro2( M, X,Y )		M(X,Y)
#define _LineString				_DoMacro1( _Stringize, __LINE__ )
#define _Indirect(A)			A

#define _StrCat(X,Y) _StrCatDelay(X,Y)
#define _StrCatDelay(X,Y) _StrCatImmediate(X,Y)
#define _StrCatImmediate(X,Y) X##Y

#define gPragmaMessage(str)		message( __FILE__ "(" _LineString ") : message: " str)
	// don't put a semicolon after a gPragmaMessage
	// use like this :
	//#pragma gPragmaMessage("my text")

#define ASSERT(exp)			do { if ( ! (exp) ) { fprintf(stderr,"%s",#exp); __asm int 3 } } while(0)

//-------------------------------------------------------------------
